#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""

The objective of this file is to help with prototyping and testing the capability state machine system.

"""
from typing import Callable, Dict, List

import pytest


class Noop:
    def __call__(self, *args, **kwargs):
        # empty because this is the no-operation action
        pass


class StartWorkflow:
    def __init__(self, workflow_name):
        self.workflow_name = workflow_name

    def __call__(self, *args, **kwargs):
        # requests.post(f"/workflows/{self.workflow_name}/create-and-submit", {})
        # ...
        pass


class QueueWorkflow:
    def __init__(self, workflow_name):
        self.workflow_name = workflow_name

    def __call__(self, *args, **kwargs):
        # message the queue manager
        pass


class Notify:
    def __init__(self, user, notification):
        self.user = user
        self.notification = notification

    def __call__(self, *args, **kwargs):
        # requests.post(f"/notifications/{self.notification}/send", users=self.user)
        # ...
        pass


class Transition:
    def __init__(self, pattern: str, action: Callable, target: "State"):
        self.pattern = pattern
        self.action = action
        self.target = target

    def matches(self, message: Dict) -> bool:
        return self.pattern.split("event_type==")[1] == message["event_type"]


class State:
    def __init__(self, name):
        self.name = name
        self.transitions: List[Transition] = []

    def link(self, pattern: str, state: "State", action: Callable[[None], None] = Noop()):
        self.transitions.append(Transition(pattern, action, state))

    def signal(self, message: Dict) -> "State":
        for transition in self.transitions:
            if transition.matches(message):
                transition.action()
                return transition.target
        return self

    def accepting_state(self) -> bool:
        return not self.transitions


class MealyMachine:
    def __init__(self, states):
        self.states = states
        self.state = None

    def start(self):
        self.state = self.states[0]
        self.signal({"event_type": "start"})

    def signal(self, message: Dict):
        self.state = self.state.perform_action(message)

    def complete(self):
        return self.state.accepting_state()


@pytest.fixture
def basic_capability() -> MealyMachine:
    start = State("Start")
    executing = State("Executing")
    failed = State("Failed")
    complete = State("Complete")

    start.link(pattern="event_type==start", action=StartWorkflow("foo"), state=executing)
    executing.link(pattern="event_type==workflow-complete", action=Notify("user", "workflow-complete"), state=complete)
    executing.link(pattern="event_type==workflow-failed", action=Notify("user", "workflow-failed"), state=failed)

    return MealyMachine([start, executing, failed, complete])


@pytest.fixture
def qa_machine() -> MealyMachine:
    # QA is a complex workflow with the following states:
    # Waiting for Approval --[Rejected]-> Rejected
    #     |
    #   [Approved]
    #     |
    #     V
    # Executing  <-[Retry]----  Waiting for QA
    #            --[Complete]->    |        ^
    #                              |        |
    #                        [Approved]   [Rejected]
    #                              |        |
    #                              V        |
    #                            Waiting for AOD    --[Accepted]->  Ingesting -[Complete]-> Complete
    #
    w_approve = State("Waiting for Approval")
    executing = State("Executing")
    w_qa = State("Waiting for QA")
    w_aod = State("Waiting for AOD")
    exec_ingest = State("Ingesting")
    failed = State("Failed")
    complete = State("Complete")

    w_approve.link(pattern="event_type==reject", action=Noop(), state=failed)
    w_approve.link(pattern="event_type==approve", action=QueueWorkflow("std_calibration"), state=executing)

    executing.link(pattern="event_type==workflow-complete", action=Notify("qa", "qa-ready-message"), state=w_qa)

    w_qa.link(pattern="event_type==retry", action=QueueWorkflow("std_calibration"), state=executing)
    w_qa.link(pattern="event_type==approve", action=Notify("aod", "qa-accept-message"), state=w_aod)

    w_aod.link(pattern="event_type==reject", action=Notify("qa", "aod-reject-message"), state=w_qa)
    w_aod.link(pattern="event_type==accept", action=StartWorkflow("ingest"), state=exec_ingest)

    exec_ingest.link(pattern="event_type==workflow-complete", action=Notify("user", "cal-ingested"), state=complete)

    return MealyMachine([w_approve, executing, w_qa, w_aod, exec_ingest, complete])


def test_basic_machine(basic_capability: MealyMachine):
    # walk through the happy path
    basic_capability.start()
    assert basic_capability.state.name == "Executing"
    basic_capability.signal({"event_type": "workflow-complete"})
    assert basic_capability.state.name == "Complete"
    assert basic_capability.complete()

    # less happy path
    basic_capability.start()
    assert basic_capability.state.name == "Executing"
    basic_capability.signal({"event_type": "workflow-failed"})
    assert basic_capability.state.name == "Failed"
    assert basic_capability.complete()


def test_qa_transitions_happy_path(qa_machine: MealyMachine):
    # so, let's first walk it through the normal flow
    qa_machine.start()
    qa_machine.signal({"event_type": "approve"})
    assert qa_machine.state.name == "Executing"
    qa_machine.signal({"event_type": "workflow-complete"})
    assert qa_machine.state.name == "Waiting for QA"
    qa_machine.signal({"event_type": "approve"})
    assert qa_machine.state.name == "Waiting for AOD"
    qa_machine.signal({"event_type": "accept"})
    assert qa_machine.state.name == "Ingesting"
    qa_machine.signal({"event_type": "workflow-complete"})
    assert qa_machine.state.name == "Complete"
    assert qa_machine.complete()


def test_qa_transitions_early_exit(qa_machine: MealyMachine):
    # now let's walk through the early termination scenario
    qa_machine.start()
    qa_machine.signal({"event_type": "reject"})
    assert qa_machine.state.name == "Failed"
    assert qa_machine.complete()


def test_qa_transitions_looping(qa_machine: MealyMachine):
    # now let's try some looping scenarios
    qa_machine.start()
    qa_machine.signal({"event_type": "approve"})
    for _ in range(4):
        assert qa_machine.state.name == "Executing"
        qa_machine.signal({"event_type": "workflow-complete"})
        assert qa_machine.state.name == "Waiting for QA"
        qa_machine.signal({"event_type": "retry"})

    # exit the loop
    assert qa_machine.state.name == "Executing"
    qa_machine.signal({"event_type": "workflow-complete"})
    assert qa_machine.state.name == "Waiting for QA"
    qa_machine.signal({"event_type": "approve"})

    # now let's do a loop between the AOD and the DA
    assert qa_machine.state.name == "Waiting for AOD"
    for _ in range(4):
        qa_machine.signal({"event_type": "reject"})
        assert qa_machine.state.name == "Waiting for QA"
        qa_machine.signal({"event_type": "approve"})
        assert qa_machine.state.name == "Waiting for AOD"

    # now let's exit that loop
    assert qa_machine.state.name == "Waiting for AOD"
    qa_machine.signal({"event_type": "accept"})
    assert qa_machine.state.name == "Ingesting"
    qa_machine.signal({"event_type": "workflow-complete"})
    assert qa_machine.state.name == "Complete"
    assert qa_machine.complete()
