#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114, C0115, C0116, E0401, R0201, W0613, W0621, W1203

from __future__ import annotations

import datetime
import json
import logging
import pathlib
import re
from typing import Dict, List, Optional

import pendulum
import requests
import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship

from workspaces.capability.enums import (
    CapabilityRequestState,
    CapabilityVersionState,
    QueueState,
)
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema_interfaces import (
    ActionIF,
    CapabilityExecutionIF,
    CapabilityIF,
    CapabilityRequestIF,
    CapabilityVersionFileIF,
    CapabilityVersionIF,
    PatternIF,
)
from workspaces.capability.services.interfaces import ExecutionManagerIF
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import AbstractFile, JSONSerializable

Base = declarative_base()

logger = logging.getLogger(__name__)

CAPABILITY_NAME_FK = "capabilities.capability_name"


class Action(Base, ActionIF):
    __tablename__ = "capability_state_actions"
    action_id = sa.Column(
        "action_id",
        sa.Integer,
        primary_key=True,
        autoincrement=True,
        comment="The unique identifier for this state action",
    )
    transition_id = sa.Column(
        "transition_id",
        sa.Integer,
        sa.ForeignKey("capability_state_transitions.transition_id"),
        nullable=False,
        comment="The transition this action is executed during",
    )
    action_type = sa.Column(
        "action_type",
        sa.String,
        nullable=False,
        comment="The type of action this is. Most likely the name of an Action subclass such as SendNotification or ExecuteWorkflow",
    )
    arguments = sa.Column("arguments", sa.String, comment="Additional arguments for the action")
    transition = relationship("Transition", back_populates="actions")
    __mapper_args__ = {"polymorphic_on": action_type}


class Noop(Action):
    __mapper_args__ = {"polymorphic_identity": "Noop"}

    def __call__(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        # empty because this is the no-operation action
        pass


class ExecuteWorkflow(Action):
    __mapper_args__ = {"polymorphic_identity": "ExecuteWorkflow"}

    def __call__(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        logger.info("Executing workflow %s for execution %s", self.arguments, execution.id)
        workflow_name = self.arguments
        workflow_args = None
        parameters = execution.version.parameters
        files = execution.version.files
        if parameters is not None:
            # convert to json string because json cares about single vs. double quotes.
            # DO NOT TAKE THIS OUT! Python will yell at you.
            parameters = json.dumps(parameters)
            workflow_args = json.loads(parameters)
        logger.info(
            "submitting workflow request with args %s (and type %s)",
            workflow_args,
            type(workflow_args),
        )
        manager.submit_workflow_request(execution.id, workflow_name, workflow_args, files)


class QueueWorkflow(Action):
    __mapper_args__ = {"polymorphic_identity": "QueueWorkflow"}

    def __call__(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        # message the execution manager
        logger.info(f"Enqueuing workflow {self.arguments}")
        manager.queue_execution(execution)


class SendNotification(Action):
    __mapper_args__ = {"polymorphic_identity": "SendNotification"}

    def __call__(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        # requests.post(f"/notifications/{self.notification}/send", users=self.user)
        # ...
        logger.info("SEND NOTIFICATION!!!")


class SendMessage(Action):
    """
    Action that simply sends an AMQP message with the type of the given arguments
    """

    __mapper_args__ = {"polymorphic_identity": "SendMessage"}

    def __call__(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        message_type = self.arguments
        message_architect = CapabilityMessageArchitect(execution=execution)
        message_to_send = message_architect.compose_message(message_type)
        logger.info("Sending %s message regarding execution #%s", message_to_send["type"], execution.id)
        manager.messenger.send_message(**message_to_send)


class QaPass(Action):
    """
    Action that performs system-level operations required by QA-passing a capability request
    """

    __mapper_args__ = {"polymorphic_identity": "QaPass"}

    def _fail_other_versions(self, request: CapabilityRequest):
        """
        Given a capability request, set the state of the current version to Complete, and the state of all other
        versions to Failed

        :param request: Capability request
        """
        versions = request.versions
        for version in versions:
            if version.__json__() == request.current_version.__json__():
                # Mark current version as complete
                print(f"Passing version: {version.__json__()}")
                version.state = CapabilityVersionState.Complete.name
            else:
                # Mark all other versions as failed
                print(f"Failing version: {version.__json__()}")
                version.state = CapabilityVersionState.Failed.name
                # Abort executions of version
                self._abort_running_executions(version)

    def _abort_running_executions(self, version: CapabilityVersion):
        """
        Given a capability version, fail all running executions

        :param version: Capability version
        """
        workflow_service_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

        for execution in version.executions:
            workflow_request_id = execution.current_workflow_request_id
            logger.info(f"Attempting to abort running execution #{execution.id}.")
            requests.post(f"{workflow_service_url}/workflows/requests/{workflow_request_id}/abort")

    def __call__(self, execution: CapabilityExecutionIF, *args):
        logger.info("Execution #%s has passed QA", execution.id)
        request = execution.capability_request
        self._fail_other_versions(request)


class AnnounceQa(Action):
    """
    Action that makes a REST call to the announce_qa endpoint in the workflow service, which announces that an execution
    is ready for QA
    """

    __mapper_args__ = {"polymorphic_identity": "AnnounceQa"}

    def __call__(self, execution: CapabilityExecutionIF, *args):
        workflow_request_id = execution.current_workflow_request_id
        msg_type = "qa_ready"
        workflow_service_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
        url = f"{workflow_service_url}/workflows/requests/{workflow_request_id}/qa/{msg_type}"
        requests.post(url)


class State(Base):
    transitions: List[Transition]

    __tablename__ = "capability_states"

    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey(CAPABILITY_NAME_FK),
        primary_key=True,
        comment="The capability this state belongs to",
    )
    state_name = sa.Column(
        "state",
        sa.String,
        nullable=False,
        primary_key=True,
        comment="A state in this state machine, such as 'Executing' or 'Awaiting QA'",
    )

    def link(self, pattern: str, state: State, action: ActionIF):
        transition = Transition(pattern=pattern, from_state=self, to_state=state)
        transition.actions.append(action)
        self.transitions.append(transition)

    def find_matching_transition(self, message: Dict) -> Optional[Transition]:
        """
        Find matching transition for state based on a given message

        :param message: Message to match transition with
        :return: Matching transition if one is found; else None
        """
        for transition in self.transitions:
            if transition.matches(message):
                return transition
        return None

    def get_next_state(self, message: Dict) -> State:
        """
        Get next state for current transition, given a message

        :param message: Message to match transition with
        :return: Next state if matching transition is found; else current state
        """
        transition = self.find_matching_transition(message)
        if transition:
            return transition.to_state
        return self

    def perform_action(self, message: Dict, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        """
        Perform action for the current transition, if one exists

        :param message: Message to match transition with
        :param execution: Execution that is performing the action
        :param manager: Execution manager to send to the action
        """
        transition = self.find_matching_transition(message)
        if transition:
            transition.action(execution, manager)

    def final(self) -> bool:
        return not self.transitions

    def __str__(self):
        return self.state_name


class Capability(Base, CapabilityIF):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """

    __tablename__ = "capabilities"
    name = sa.Column("capability_name", sa.String, primary_key=True)
    start_state = sa.Column(
        "start_state",
        sa.String,
        # Magic string required here, I think  -nhertz
        default="Start",
        server_default="Start",
    )
    max_jobs = sa.Column("max_jobs", sa.Integer)
    enabled = sa.Column("enabled", sa.Boolean, default=True, server_default="true")
    paused = sa.Column("paused", sa.Boolean, default=False, server_default="false")
    has_image_products = sa.Column(
        "has_image_products", sa.Boolean, default=False, server_default="false", nullable=False
    )
    requests = relationship("CapabilityRequest", back_populates="capability")

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF):
        """
        Create a new request for this capability with specific options

        FIXME: Future products needs a string representation
        """
        request = CapabilityRequest(capability=self, parameters=str(parameters))
        self.requests.append(request)
        return request

    def __str__(self) -> str:
        return f"Capability: {self.name}, max concurrent jobs of {self.max_jobs}"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "name": self.name,
            "max_jobs": self.max_jobs,
            "enabled": self.enabled,
            "has_image_products": self.has_image_products,
            "paused": self.paused,
        }

    @classmethod
    def from_json(cls, json: dict) -> Capability:
        """
        Method that creates an instance of a Capability from given JSON

        :param json: the description of this capabiity in a dictionary
        :return: Capability instance
        """
        copy = dict(json)
        del copy["type"]
        return Capability(**copy)


class Transition(Base):

    pattern: PatternIF
    actions: List[ActionIF]
    from_state: State
    to_state: State

    __tablename__ = "capability_state_transitions"
    transition_id = sa.Column(
        "transition_id",
        sa.Integer,
        primary_key=True,
        autoincrement=True,
        comment="The unique identifier of this state machine transition",
    )

    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey(Capability.name),
        nullable=False,
        comment="The capability to whose state machine this transition belongs",
    )

    from_state_name = sa.Column(
        "from_state",
        sa.String,
        nullable=False,
        comment="The starting state for this transition",
    )

    pattern = sa.Column(
        "pattern",
        sa.String,
        nullable=False,
        comment="A JSONPATH pattern for what events cause us to take this transition",
    )

    to_state_name = sa.Column(
        "to_state",
        sa.String,
        nullable=False,
        comment="The destination state for this transition",
    )

    from_state = relationship(State, foreign_keys=[capability_name, from_state_name], backref="transitions")
    to_state = relationship(State, foreign_keys=[capability_name, to_state_name])

    actions = relationship(Action, back_populates="transition")

    __table_args__ = (
        sa.ForeignKeyConstraint(
            (capability_name, from_state_name),
            [State.capability_name, State.state_name],
        ),
        sa.ForeignKeyConstraint((capability_name, to_state_name), [State.capability_name, State.state_name]),
        {},
    )

    def matches(self, message: Dict) -> bool:
        """
        Check if a message matches a given pattern

        Patterns are stored (in the database) in the following format:
        {key] == {value}

        Only one pattern per transition is supported currently

        :param message: Message to be checked for matches
        :return: True if there's a match; else False
        """
        key, value, *rest = re.split(r"\s*==\s*", self.pattern)
        if len(rest) or not value:
            logger.info(
                f"Pattern {self.pattern} is malformed. It should take the form '{{key}} == {{value}}', "
                f"corresponding to the key and value of the incoming message."
            )
        logger.info(
            "Checking to see if key %s has value %s (it is %s)", key, value, message[key] if key in message else "None"
        )
        return key in message.keys() and value == message[key]

    def action(self, execution: CapabilityExecutionIF, manager: ExecutionManagerIF):
        logger.info("Processing my actions")
        for action in self.actions:
            logger.info("Processing an action of type %s", action.action_type)
            action(execution, manager)


class CapabilityRequest(Base, CapabilityRequestIF):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """

    __tablename__ = "capability_requests"
    id = sa.Column("capability_request_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))
    ingested = sa.Column("ingested", sa.Boolean, default=False, server_default="false", nullable=False)
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )

    versions = relationship(
        "CapabilityVersion",
        back_populates="request",
        order_by="CapabilityVersion.version_number",
        collection_class=ordering_list("version_number"),
    )
    capability = relationship(Capability, back_populates="requests")

    def update_state(self, state: CapabilityRequestState):
        self.state = state.name

    def determine_state(self):
        """
        Determine state of request based on the state of its versions and set it accordingly

        RULES:
        - If there is a complete version, the request is complete
        - If all versions are failed, the request is failed
        - If all versions are created, the request is created
        - Otherwise, it is submitted
        """
        logger.info("Determine request state...")
        logger.info(f"Current version: {self.current_version.__json__()}")
        version_states = [version.state for version in self.versions]

        if self.current_version.state == CapabilityVersionState.Complete.name:
            # The current version is complete, so the request is complete
            logger.info("Current version is complete. Setting request to Complete.")
            self.state = CapabilityRequestState.Complete.name
        elif all(state == CapabilityRequestState.Failed.name for state in version_states):
            # Request has all failed versions, so it is failed
            logger.info("All versions are failed. Setting request to Failed.")
            self.state = CapabilityRequestState.Failed.name
        elif all(state == CapabilityRequestState.Created.name for state in version_states):
            # Request has no submitted versions, so it is still in the created state
            logger.info("All versions are created. Setting request to Created.")
            self.state = CapabilityRequestState.Created.name
        else:
            logger.info("Versions are in a mixture of states. Setting request to Submitted.")
            self.state = CapabilityRequestState.Submitted.name

    def __str__(self):
        return f"CapabilityRequest object: {self.__dict__}"

    @property
    def created_at_ago(self) -> str:
        return pendulum.instance(self.created_at).diff_for_humans()

    @property
    def updated_at_ago(self) -> str:
        return pendulum.instance(self.updated_at).diff_for_humans()

    def __repr__(self):
        return (
            f"<CapabilityRequest #{self.id} "
            f"for capability {self.capability_name}, "
            f"created {self.created_at_ago} and "
            f"last updated {self.updated_at_ago})>"
        )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "id": self.id,
            "capability_name": self.capability_name,
            "state": self.state,
            "parameters": self.current_version.parameters if self.current_version else None,
            "ingested": self.ingested,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "current_execution": self.current_execution.__json__() if self.current_execution else None,
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        del copy["parameters"]
        del copy["current_execution"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityRequest(**copy)

    @property
    def current_version(self) -> CapabilityVersion:
        return self.versions[-1] if len(self.versions) else None

    @property
    def current_execution(self) -> CapabilityExecution:
        return self.current_version.current_execution if self.current_version else None


class CapabilityVersion(Base, CapabilityVersionIF):
    """
    Schema representation of a capability version
    """

    __tablename__ = "capability_versions"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
        primary_key=True,
    )
    version_number = sa.Column("version_number", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String, default="Created", server_default="Created")
    parameters = sa.Column("parameters", sa.JSON)
    workflow_metadata = sa.Column("workflow_metadata", sa.JSON)
    request = relationship(CapabilityRequest, back_populates="versions")
    executions = relationship("CapabilityExecution", back_populates="version")
    files = relationship("CapabilityVersionFile", back_populates="version")
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))
    capability = relationship(Capability)

    @property
    def current_execution(self) -> CapabilityExecution:
        return self.executions[-1] if len(self.executions) else None

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "capability_request_id": self.capability_request_id,
            "version_number": self.version_number,
            "state": self.state,
            "parameters": self.parameters,
            "workflow_metadata": self.workflow_metadata,
            "files": [file.__json__() for file in self.files],
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        del copy["files"]
        return CapabilityVersion(**copy)


class CapabilityVersionFile(Base, CapabilityVersionFileIF):
    """
    A Capability Request File is a file that was supplied by the user and attached to the current version
    """

    __tablename__ = "capability_version_files"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        primary_key=True,
    )
    version_number = sa.Column(
        "version_number",
        sa.Integer,
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True, nullable=False)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    version = relationship(CapabilityVersion, back_populates="files")

    __table_args__ = (
        sa.ForeignKeyConstraint(
            [capability_request_id, version_number],
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
        {},
    )

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<CapabilityVersionFile filename={self.filename}>"

    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "capability_request_id": self.capability_request_id,
            "version_number": self.version_number,
            "filename": self.filename,
            "content": self.content.hex(),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return CapabilityVersionFile(filename=json["filename"], content=bytes.fromhex(json["content"]))


class SaPath(sa.types.TypeDecorator):
    impl = sa.types.String

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return pathlib.Path(value)

    def copy(self, **kw):
        return SaPath(self.impl.length)


class CapabilityExecution(Base, CapabilityExecutionIF, JSONSerializable):
    """
    Schema representation of a capability request's execution record
    """

    __tablename__ = "capability_executions"

    id = sa.Column("execution_id", sa.Integer, primary_key=True)
    state_name = sa.Column("state", sa.String)
    queue_state = sa.Column(
        "queue_state", sa.String, default=QueueState.NotQueued.name, server_default=QueueState.NotQueued.name
    )
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
    )
    version_number = sa.Column("capability_version_number", sa.Integer)
    version = relationship(CapabilityVersion, back_populates="executions")
    current_workflow_request_id = sa.Column("current_workflow_request_id", sa.Integer)
    delivery_url = sa.Column("delivery_url", sa.String)
    delivery_path = sa.Column("delivery_path", SaPath)
    capability_name = sa.Column("capability_name", sa.String, sa.ForeignKey(CAPABILITY_NAME_FK))

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )
    capability = relationship(Capability, viewonly=True)
    state = relationship(State)

    __table_args__ = (
        sa.ForeignKeyConstraint((capability_name, state_name), [State.capability_name, State.state_name]),
        sa.ForeignKeyConstraint(
            (capability_request_id, version_number),
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
    )

    @property
    def capability_request(self):
        return self.version.request

    def __lt__(self, other):
        return self.created_at < other.created_at

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return dict(
            type=self.__class__.__name__,
            id=self.id,
            state_name=self.state_name,
            queue_state=self.queue_state,
            capability_request_id=self.capability_request_id,
            version_number=self.version_number,
            current_workflow_request_id=self.current_workflow_request_id,
            delivery_url=self.delivery_url,
            created_at=self.created_at.isoformat(),
            updated_at=self.updated_at.isoformat(),
        )

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityExecution(**copy)

    def signal(self, manager: ExecutionManagerIF, message: Dict):
        logger.info("Signalling my state %s with this message of type %s", self.state_name, message["type"])
        previous_state = self.state
        # Transition execution state
        self.state = self.state.get_next_state(message)
        # Perform action
        previous_state.perform_action(message, self, manager)
